# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import capellambse

# pylint: disable=pointless-statement


def test_AttributeAuditor_tracks_uuids_of_accessed_ModelElements(
    model: capellambse.MelodyModel,
):
    comp, comp1, comp2 = model.la.all_components[:3]
    expected = {comp.uuid, comp2.uuid}
    with capellambse.AttributeAuditor(
        model, {"name", "description"}
    ) as recorded_ids:
        comp.name
        comp1.uuid
        comp2.description

    assert expected == recorded_ids


def test_AttributeAuditor_handles_recursive_uuid(
    model: capellambse.MelodyModel,
):
    comp = model.la.all_components[0]
    with capellambse.AttributeAuditor(model, {"uuid"}) as recorded_ids:
        comp.uuid

    assert {comp.uuid} == recorded_ids


def test_AttributeAuditor_destroys_model_reference_when_detach(
    model: capellambse.MelodyModel,
):
    auditor = capellambse.AttributeAuditor(model, {"name"})

    auditor.detach()

    assert auditor.model is None



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Global fixtures for pytest."""
import io
import pathlib
import sys

import pytest

import capellambse

INSTALLED_PACKAGE = pathlib.Path(capellambse.__file__).parent

TEST_ROOT = pathlib.Path(__file__).parent / "data" / "melodymodel"
TEST_MODEL = "Melody Model Test.aird"


@pytest.fixture(scope="session")
def session_shared_model() -> capellambse.MelodyModel:
    """Load the standard test model.

    Unlike the ``model`` fixture, this fixture is shared across the
    entire test session. As such, the test functions using this fixture
    are expected to not modify it.

    This fixture exists as a speed optimization for tests that only read
    from the model.
    """
    return capellambse.MelodyModel(TEST_ROOT / "5_0" / TEST_MODEL)


@pytest.fixture
def model(monkeypatch) -> capellambse.MelodyModel:
    """Return the Capella 5.0 test model."""
    monkeypatch.setattr(sys, "stderr", io.StringIO)
    return capellambse.MelodyModel(TEST_ROOT / "5_0" / TEST_MODEL)


@pytest.fixture
def model_5_2(monkeypatch) -> capellambse.MelodyModel:
    """Return the Capella 5.2 test model."""
    monkeypatch.setattr(sys, "stderr", io.StringIO)
    return capellambse.MelodyModel(TEST_ROOT / "5_2" / TEST_MODEL)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Implements a high-level interface to Capella projects."""
from __future__ import annotations

__all__ = [
    "ElementList",
    "GenericElement",
    "NonUniqueMemberError",
    "MelodyModel",
]

import collections.abc as cabc
import logging
import os
import pathlib
import typing as t

from lxml import etree

import capellambse
import capellambse.helpers
import capellambse.pvmt
from capellambse import filehandler, loader
from capellambse.loader import xmltools

from . import common, diagram  # isort:skip

# Architectural Layers
from .layers import oa, ctx, la, pa  # isort:skip

# Exports
from .common import (  # isort:skip
    ElementList,
    GenericElement,
    NonUniqueMemberError,
)

LOGGER = logging.getLogger(__name__)
XT_PROJECT = "org.polarsys.capella.core.data.capellamodeller:Project"
XT_LIBRARY = "org.polarsys.capella.core.data.capellamodeller:Library"
XT_SYSENG = "org.polarsys.capella.core.data.capellamodeller:SystemEngineering"


class MelodyModel:
    """Provides high-level access to a model.

    This class builds upon the lower-level
    :class:`~capellambse.loader.core.MelodyLoader` to provide an
    abstract, high-level interface for easy access to various model
    aspects.
    """

    oa = common.DirectProxyAccessor(oa.OperationalAnalysis, rootelem=XT_SYSENG)
    sa = common.DirectProxyAccessor(ctx.SystemAnalysis, rootelem=XT_SYSENG)
    la = common.DirectProxyAccessor(la.LogicalArchitecture, rootelem=XT_SYSENG)
    pa = common.DirectProxyAccessor(
        pa.PhysicalArchitecture, rootelem=XT_SYSENG
    )
    diagrams = diagram.DiagramAccessor(
        None, cacheattr="_MelodyModel__diagram_cache"
    )

    uuid = xmltools.AttributeProperty(
        "_element",
        "id",
        writable=False,
        __doc__="The unique ID of the model's root element.",
    )
    name = xmltools.AttributeProperty(
        "_element", "name", __doc__="The name of this model."
    )

    _diagram_cache: filehandler.FileHandler
    _diagram_cache_subdir: pathlib.PurePosixPath

    def __init__(
        self,
        path: str | os.PathLike,
        *,
        diagram_cache: (
            str
            | os.PathLike
            | filehandler.FileHandler
            | dict[str, t.Any]
            | None
        ) = None,
        diagram_cache_subdir: str | pathlib.PurePosixPath | None = None,
        jupyter_untrusted: bool = False,
        **kwargs: t.Any,
    ) -> None:
        # pylint: disable=line-too-long
        """Load a project.

        For complete information on which exact ``kwargs`` are
        supported, consult the documentation of the used file handler.
        Refer to the "See Also" section for a collection of links.

        Below are some common parameter names and their usual meanings.
        Not all file handlers support all parameters.

        .. note::
            Passing in arguments that are not accepted by the selected
            file handler will result in an exception being raised.
            Similarly, leaving out arguments that are required by the
            file handler will also result in an exception.

        Parameters
        ----------
        path
            Path or URL to the project. The following formats are
            accepted:

            * A path to a local ``.aird`` file.
            * A path to a local directory (requires ``entrypoint``).
            * An SCP-style short URL, which will be treated as referring
              to a Git repository.

              Example: ``git@github.com:DSD-DBS/py-capellambse.git``
            * A remote URL, with a protocol or prefix that indicates
              which file handler to invoke (requires ``entrypoint``).

              Some examples:

              * ``git://git.example.com/model/coffeemaker.git``
              * ``git+https://git.example.com/model/coffeemaker.git``
              * ``git+ssh://git@git.example.com/model/coffeemaker.git``

              .. note:: Depending on the exact file handler, saving back
                 to a remote location might fail with ``update_cache``
                 set to ``False``. See :meth:`save` for more details.
        entrypoint: str
            Entrypoint from path to the main ``.aird`` file.
        revision: str
            The revision to use, if loading a model from a version
            control system like git. Defaults to the current HEAD. If
            the used VCS does not have a notion of "current HEAD", this
            argument is mandatory.
        disable_cache: bool
            Disable local caching of remote content.
        update_cache: bool
            Update the local cache. Defaults to ``True``, but can be
            disabled to reuse the last cached state.
        identity_file: str | pathlib.Path
            The identity file (private key) to use when connecting via
            SSH.
        known_hosts_file: str | pathlib.Path
            The ``known_hosts`` file to pass to SSH for verifying the
            server's host key.
        username: str
            The username to log in as remotely.
        password: str
            The password to use for logging in. Will be ignored when
            ``identity_file`` is passed as well.
        diagram_cache: str | pathlib.Path | ~capellambse.filehandler.FileHandler | dict[str, ~typing.Any]
            An optional place where to find pre-rendered, cached
            diagrams. When a diagram is found in this cache, it will be
            loaded from there instead of being rendered on access. Note
            that diagrams will only be loaded from there, but not be put
            back, i.e. to use it effectively, the cache has to be
            pre-populated.

            This argument accepts the following values:

            - ``None``, in which case all diagrams will be rendered
              internally the first time they are used.
            - A path to a local directory.
            - A URL just like for the ``path`` argument.
            - A dictionary with the arguments to a
              :py:class:`~capellambse.FileHandler`. The dict's ``path``
              key will be analyzed to determine the correct FileHandler
              class.
            - An instance of :py:class:`~capellambse.FileHandler`, which
              will be used directly.

            .. warning:: When using the diagram cache, always make sure
               that the cached diagrams actually match the model version
               that is being used. There is no way to check this
               automatically.

            The file names looked up in the cache built in the format
            ``uuid.ext``, where ``uuid`` is the UUID of the diagram (as
            reported by ``diag_obj.uuid``) and ``ext`` is the render
            format. Example:

            - Diagram ID: ``_7FWu4KrxEeqOgqWuHJrXFA``
            - Render call: ``diag_obj.as_svg`` or ``diag_obj.render("svg")``
            - Cache file name: ``_7FWu4KrxEeqOgqWuHJrXFA.svg``

            *This argument is **not** passed to the file handler.*
        diagram_cache_subdir: str
            A sub-directory prefix to prepend to diagram UUIDs before
            looking them up in the ``diagram_cache``.

            *This argument is **not** passed to the file handler.*
        jupyter_untrusted: bool
            If set to True, restricts or disables some features that are
            unavailable in an untrusted Jupyter environment. Currently
            this only disables the SVG format as rich display option for
            Ipython, which is needed to avoid rendering issues with
            Github's Jupyter notebook viewer.

        See Also
        --------
        capellambse.filehandler.FileHandler :
            Abstract super class for file handlers. Contains information
            needed for implementing custom handlers.
        capellambse.filehandler.localfilehandler.LocalFileHandler :
            The file handler responsible for local files and
            directories.
        capellambse.filehandler.gitfilehandler.GitFileHandler :
            The file handler implementing the ``git://`` protocol.
        capellambse.filehandler.http.HTTPFileHandler :
            A simple ``http(s)://`` file handler.
        """
        # pylint: enable=line-too-long
        capellambse.load_model_extensions()

        self._loader = loader.MelodyLoader(path, **kwargs)
        self.info = self._loader.get_model_info()
        self.jupyter_untrusted = jupyter_untrusted

        try:
            self._pvext = capellambse.pvmt.load_pvmt_from_model(self._loader)
        except ValueError as err:
            LOGGER.warning(
                "Cannot load PVMT extension: %s: %s", type(err).__name__, err
            )
            LOGGER.warning("Property values are not available in this model")
            self._pvext = None

        if diagram_cache:
            if diagram_cache == path:
                self._diagram_cache = self._loader.filehandler
            elif isinstance(diagram_cache, filehandler.FileHandler):
                self._diagram_cache = diagram_cache
            elif isinstance(diagram_cache, cabc.Mapping):
                self._diagram_cache = filehandler.get_filehandler(
                    **diagram_cache
                )
            else:
                self._diagram_cache = filehandler.get_filehandler(
                    diagram_cache
                )
            self._diagram_cache_subdir = pathlib.PurePosixPath(
                diagram_cache_subdir or "/"
            )

    @property
    def _element(self) -> etree._Element:
        for tree in self._loader.trees.values():
            if capellambse.helpers.xtype_of(tree.root) in {
                XT_PROJECT,
                XT_LIBRARY,
            }:
                return tree.root
        raise TypeError("No viable root element found")

    @property
    def _model(self) -> MelodyModel:
        return self

    def save(self, **kw: t.Any) -> None:
        # pylint: disable=line-too-long
        """Save the model back to where it was loaded from.

        Parameters
        ----------
        kw
            Additional keyword arguments accepted by the file handler in
            use. Please see the respective documentation for more info.

        See Also
        --------
        capellambse.filehandler.localfilehandler.LocalFileHandler.write_transaction :
            Accepted ``**kw`` when using local directories
        capellambse.filehandler.gitfilehandler.GitFileHandler.write_transaction :
            Accepted ``**kw`` when using ``git://`` and similar URLs

        Notes
        -----
        With a file handler that contacts a remote location (such as the
        :class:`~capellambse.filehandler.gitfilehandler.GitFileHandler`
        with non-local repositories), saving might fail if the local
        state has gone out of sync with the remote state. To avoid this,
        always leave the ``update_cache`` parameter at its default value
        of ``True`` if you intend to save changes.
        """
        # pylint: enable=line-too-long
        self._loader.save(**kw)

    def search(
        self,
        *xtypes: str | type[common.GenericElement],
        below: common.GenericElement | None = None,
    ) -> common.ElementList:
        r"""Search for all elements with any of the given ``xsi:type``\ s.

        If only one xtype is given, the return type will be
        :class:`common.ElementList`, otherwise it will be
        :class:`common.MixedElementList`.

        If no ``xtypes`` are given at all, this method will return an
        exhaustive list of all (semantic) model objects that have an
        ``xsi:type`` set.

        Parameters
        ----------
        xtypes
            The ``xsi:type``\ s to search for, or the classes
            corresponding to them (or a mix of both).
        below
            A model element to constrain the search. If given, only
            those elements will be returned that are (immediate or
            nested) children of this element. This option takes into
            account model fragmentation, but it does not treat link
            elements specially.
        """
        xtypes_: list[str] = []
        for i in xtypes:
            if isinstance(i, type) and issubclass(i, common.GenericElement):
                xtypes_.append(common.build_xtype(i))
            elif ":" in i:
                xtypes_.append(i)
            else:
                suffix = ":" + i
                matching_types: list[str] = []
                for l in common.XTYPE_HANDLERS.values():
                    matching_types.extend(t for t in l if t.endswith(suffix))
                if not matching_types:
                    raise ValueError(f"Unknown incomplete type name: {i}")
                xtypes_.extend(matching_types)

        cls = (common.MixedElementList, common.ElementList)[len(xtypes_) == 1]
        trees = {
            k
            for k, v in self._loader.trees.items()
            if v.fragment_type is loader.FragmentType.SEMANTIC
        }
        matches = self._loader.iterall_xt(*xtypes_, trees=trees)
        if below is not None:
            matches = (
                i for i in matches if below._element in i.iterancestors()
            )
        return cls(self, list(matches), common.GenericElement)

    def by_uuid(self, uuid: str) -> common.GenericElement:
        """Search the entire model for an element with the given UUID."""
        return common.GenericElement.from_model(self, self._loader[uuid])



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""The capellambse package."""
import platformdirs

dirs = platformdirs.PlatformDirs("capellambse")
del platformdirs

from importlib import metadata

try:
    __version__ = metadata.version("capellambse")
except metadata.PackageNotFoundError:
    __version__ = "0.0.0+unknown"
del metadata

from ._namespaces import *
from .auditing import AttributeAuditor
from .cli_helpers import *
from .filehandler import *
from .model import MelodyModel
from .model.common import ModelObject

_has_loaded_extensions = False


def load_model_extensions() -> None:
    """Load all model extensions.

    This function loads all entry points in the group
    ``capellambse.model_extensions`` and executes them.

    Note that this function must be placed at the end of the top-level
    ``__init__.py``, in order to ensure that all submodules were
    initialized before loading any extensions.
    """
    # pylint: disable=import-outside-toplevel  # Reduce namespace pollution
    import importlib.metadata as imm  # pylint: disable=reimported

    # pylint: disable=redefined-outer-name  # false-positive
    import logging
    import sys

    global _has_loaded_extensions
    if _has_loaded_extensions:
        return
    _has_loaded_extensions = True

    if sys.version_info < (3, 10):
        try:
            entrypoints = imm.entry_points()["capellambse.model_extensions"]
        except KeyError:
            return
    else:
        entrypoints = imm.entry_points(group="capellambse.model_extensions")

    for entrypoint in entrypoints:
        try:
            initfunc = entrypoint.load()
            initfunc()
        except Exception:
            logging.getLogger(__name__).exception(
                "Cannot load model extension %r from %r",
                entrypoint,
                entrypoint.value,
            )



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from capellambse import MelodyModel


def test_function_port_has_state_machines(model: MelodyModel):
    port = model.by_uuid("226878bc-3e4b-4236-ba72-996fc1d988c0")
    stm = model.by_uuid("06cefb2b-534e-4453-9aba-fe53329197ad")
    assert stm in port.state_machines



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

__all__ = ["UnsupportedPluginError", "UnsupportedPluginVersionError"]

import collections.abc as cabc
import dataclasses
import re
import types
import typing as t

PLUGIN_PATTERN = re.compile(r"(.*?)(\d\.\d\.\d)?$")


class UnsupportedPluginError(ValueError):
    """Raised when a plugin is unknown."""


class UnsupportedPluginVersionError(ValueError):
    """Raised when the plugin's version is unsupported."""

    def __str__(self) -> str:
        plugin, ns_plugin = self.args
        assert isinstance(plugin, Plugin) and isinstance(ns_plugin, Plugin)
        return (
            f"Plugin '{plugin.name}' with unsupported version encountered: "
            f"{plugin.version!r} does not match {ns_plugin.version!r}."
        )


@dataclasses.dataclass(frozen=True)
class Plugin:
    """Capella xml-element/plugin with info about name and version."""

    name: str
    version: str | tuple[str, str] | None = None

    @property
    def _values(self) -> cabc.Iterator[str | None]:
        yield self.name
        if isinstance(self.version, tuple):
            min_supported_version, max_supported_version = self.version
            yield min_supported_version
            yield max_supported_version
        else:
            yield self.version

    @property
    def has_version(self) -> bool:
        return self.version is not None

    def __le__(self, other: Plugin) -> bool:
        if self.version is None:
            raise AttributeError(f"Plugin '{self.name}' has no version")

        if isinstance(other.version, tuple):
            assert all((isinstance(o, str) for o in other.version))
            vmin, vmax = tuple(map(_tofloat, other.version))
            if isinstance(self.version, tuple):
                vsmin, vsmax = tuple(map(_tofloat, self.version))
                return vsmin <= vmin <= vmax <= vsmax
            else:
                return vmin <= _tofloat(self.version) <= vmax

        assert other.version is not None
        oversion = _tofloat(other.version)
        if isinstance(self.version, tuple):
            vsmin, vsmax = tuple(map(_tofloat, self.version))
            return vsmin <= oversion <= vsmax

        return oversion <= _tofloat(self.version)

    def __str__(self) -> str:
        if isinstance(suffix := list(self._values)[1], str):
            return self.name + suffix
        return self.name

    def __repr__(self) -> str:
        if len(_values := list(self._values)) > 2:
            name, minv, maxv = _values
            values = [name, f"[{minv}|{maxv}]"]
        else:
            name, version = _values
            values = [name]
            if version is not None:
                values.append(version)
        return "".join(values)  # type: ignore[arg-type]


def _tofloat(other: str) -> float:
    """Change 1.x.y...str into 1.x float."""
    version = other.split(".")
    if len(version) > 1:
        version = version[:2]

    return float(".".join(version))


def get_keys_and_plugins_from_namespaces_by_url(
    url: str,
) -> tuple[str, Plugin]:
    """Return key and :class:`Plugin` pair matching from ``NAMESPACES``.

    Parameters
    ----------
    url
        An ``xsi:type``-ish string of an xml-element/plugin

    Raises
    ------
    UnsupportedPluginError
        If requested ``url`` is not found in :class:`NAMESPACES`.

    Returns
    -------
    nskey, plugin
        A namespace-key string and a Plugin
    """
    match = PLUGIN_PATTERN.match(url)
    version = None
    if match is not None:
        plugin_name = match.group(1)
        if match.group(2) is not None:
            version = match.group(2)

    matched_plugins = [
        (nskey, Plugin(plugin_name, version))
        for nskey, nsplugin in NAMESPACES_PLUGINS.items()
        if plugin_name == nsplugin.name
    ]
    if not matched_plugins:
        plugin = Plugin(plugin_name, version)
        raise UnsupportedPluginError(f"{plugin}")
    assert (
        len(matched_plugins) == 1
    ), f"Ambiguous namespace {url!r}: {matched_plugins}"
    return matched_plugins[0]


def check_plugin(nskey: str, plugin: Plugin) -> None:
    r"""Check if the given ``plugin`` is supported.

    It is assumed that ``plugin`` is known by
    :class:`NAMESPACES_PLUGINS` from the ``nskey``. This can be ensured
    by a previous execution of
    :func:`get_keys_and_plugins_from_namespaces_by_url`.

    Parameters
    ----------
    nskey
        A key corresponding to given ``plugin`` to find a similiar
        plugin in :class:`NAMESPACES_PLUGINS`
    plugin
        A plugin for which a similiar plugin can be found in
        :class:`NAMESPACES_PLUGINS` from the given ``nskey``

    Raises
    ------
    UnsupportedPluginError
        If given ``plugin`` is unknown. This is the case when either:
          * It is not in :class:`NAMESPACES_PLUGINS`'s ``values`` or
          * an unknown ``nskey`` requested a plugin, i.e. ``nskey`` is
            not in :class:`NAMESPACES_PLUGINS`'s keys.
    UnsupportedPluginVersionError
        If given plugin is versioned and one of the following conditions
        is met:
          * If given ``plugin``'s version is singular and exceeds plugin
            version from :class:`NAMESPACES_PLUGINS` or
          * if ``plugin``\ s version is a range and is not contained in
            plugin version range from :class:`NAMESPACES_PLUGINS`.
    """
    try:
        my_plugin = NAMESPACES_PLUGINS[nskey]
    except KeyError as err:
        raise UnsupportedPluginError(str(plugin)) from err

    if plugin.has_version and not plugin <= my_plugin:
        raise UnsupportedPluginVersionError(plugin, my_plugin)


#: These XML namespaces are defined in every ``.aird`` document
NAMESPACES_PLUGINS: t.Final[
    cabc.Mapping[str, Plugin]
] = types.MappingProxyType(
    {
        "CapellaRequirements": Plugin(
            "http://www.polarsys.org/capella/requirements"
        ),
        "Requirements": Plugin(
            "http://www.polarsys.org/kitalpha/requirements"
        ),
        "concern": Plugin(
            "http://www.eclipse.org/sirius/diagram/description/concern/",
            "1.1.0",
        ),
        "description": Plugin(
            "http://www.eclipse.org/sirius/description/", "1.1.0"
        ),
        "description_1": Plugin(
            "http://www.eclipse.org/sirius/diagram/description/", "1.1.0"
        ),
        "description_2": Plugin(
            "http://www.eclipse.org/sirius/table/description/", "1.1.0"
        ),
        "description_3": Plugin(
            "http://www.eclipse.org/sirius/diagram/sequence/description/",
            "2.0.0",
        ),
        "diagram": Plugin("http://www.eclipse.org/sirius/diagram/", "1.1.0"),
        "diagramstyler": Plugin("http://thalesgroup.com/mde/melody/ordering"),
        "ecore": Plugin("http://www.eclipse.org/emf/2002/Ecore"),
        "filter": Plugin(
            "http://www.eclipse.org/sirius/diagram/description/filter/",
            "1.1.0",
        ),
        "libraries": Plugin(
            "http://www.polarsys.org/capella/common/libraries/",
            ("5.0.0", "6.0.0"),
        ),
        "metadata": Plugin(
            "http://www.polarsys.org/kitalpha/ad/metadata/", "1.0.0"
        ),
        "notation": Plugin(
            "http://www.eclipse.org/gmf/runtime/1.0.2/notation"
        ),
        "org.polarsys.capella.core.data.capellacommon": Plugin(
            "http://www.polarsys.org/capella/core/common/", ("5.0.0", "6.0.0")
        ),
        "org.polarsys.capella.core.data.capellacore": Plugin(
            "http://www.polarsys.org/capella/core/core/", ("5.0.0", "6.0.0")
        ),
        "org.polarsys.capella.core.data.capellamodeller": Plugin(
            "http://www.polarsys.org/capella/core/modeller/",
            ("5.0.0", "6.0.0"),
        ),
        "org.polarsys.capella.core.data.cs": Plugin(
            "http://www.polarsys.org/capella/core/cs/", ("5.0.0", "6.0.0")
        ),
        "org.polarsys.capella.core.data.ctx": Plugin(
            "http://www.polarsys.org/capella/core/ctx/", ("5.0.0", "6.0.0")
        ),
        "org.polarsys.capella.core.data.epbs": Plugin(
            "http://www.polarsys.org/capella/core/epbs/", ("5.0.0", "6.0.0")
        ),
        "org.polarsys.capella.core.data.fa": Plugin(
            "http://www.polarsys.org/capella/core/fa/", ("5.0.0", "6.0.0")
        ),
        "org.polarsys.capella.core.data.information": Plugin(
            "http://www.polarsys.org/capella/core/information/",
            ("5.0.0", "6.0.0"),
        ),
        "org.polarsys.capella.core.data.information.datatype": Plugin(
            "http://www.polarsys.org/capella/core/information/datatype/",
            ("5.0.0", "6.0.0"),
        ),
        "org.polarsys.capella.core.data.information.datavalue": Plugin(
            "http://www.polarsys.org/capella/core/information/datavalue/",
            ("5.0.0", "6.0.0"),
        ),
        "org.polarsys.capella.core.data.interaction": Plugin(
            "http://www.polarsys.org/capella/core/interaction/",
            ("5.0.0", "6.0.0"),
        ),
        "org.polarsys.capella.core.data.la": Plugin(
            "http://www.polarsys.org/capella/core/la/", ("5.0.0", "6.0.0")
        ),
        "org.polarsys.capella.core.data.oa": Plugin(
            "http://www.polarsys.org/capella/core/oa/", ("5.0.0", "6.0.0")
        ),
        "org.polarsys.capella.core.data.pa": Plugin(
            "http://www.polarsys.org/capella/core/pa/", ("5.0.0", "6.0.0")
        ),
        "re": Plugin("http://www.polarsys.org/capella/common/re/", "1.3.0"),
        "sequence": Plugin(
            "http://www.eclipse.org/sirius/diagram/sequence/", "2.0.0"
        ),
        "style": Plugin(
            "http://www.eclipse.org/sirius/diagram/description/style/", "1.1.0"
        ),
        "table": Plugin("http://www.eclipse.org/sirius/table/", "1.1.0"),
        "viewpoint": Plugin("http://www.eclipse.org/sirius/", "1.1.0"),
        "xmi": Plugin("http://www.omg.org/XMI"),
        "xsi": Plugin("http://www.w3.org/2001/XMLSchema-instance"),
    }
)
NAMESPACES: t.Final[cabc.Mapping[str, str]] = {
    nskey: str(plugin) for nskey, plugin in NAMESPACES_PLUGINS.items()
}



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import dataclasses


@dataclasses.dataclass
class ModelInfo:
    branch: str | None = None
    title: str | None = None
    url: str | None = None
    revision: str | None = None
    rev_hash: str | None = None
    capella_version: str | None = None



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Functions for parsing and interacting with diagrams in a Capella model."""
from __future__ import annotations

__all__ = [
    "DiagramDescriptor",
    "ActiveFilters",
    "enumerate_diagrams",
    "parse_diagrams",
    "parse_diagram",
]

import collections.abc as cabc
import pathlib
import typing as t
import urllib.parse

from lxml import etree

import capellambse._namespaces as _n
from capellambse import diagram, helpers, loader

from . import _common as C
from . import _filters, _semantic, _visual
from ._filters import ActiveFilters

DIAGRAM_ROOTS = {
    f"{{{_n.NAMESPACES['sequence']}}}SequenceDDiagram",
    f"{{{_n.NAMESPACES['diagram']}}}DSemanticDiagram",
}
"""Representations whose root element has one of these tags are diagrams.

Other representations (e.g. data tables) will not be listed by
`enumerate_diagrams`.
"""


class DiagramDescriptor(t.NamedTuple):
    fragment: pathlib.PurePosixPath
    name: str
    styleclass: str | None
    descriptor: etree._Element
    uid: str
    viewpoint: str
    target: etree._Element


def enumerate_diagrams(
    model: loader.MelodyLoader,
) -> cabc.Iterator[DiagramDescriptor]:
    """Enumerate the diagrams in the model.

    Parameters
    ----------
    model
        The MelodyLoader instance
    """
    raw_views = model.xpath2(C.XP_VIEWS)
    views: list[tuple[pathlib.PurePosixPath, etree._Element, str]] = []
    if len(raw_views) == 0:
        raise ValueError("Invalid XML: No viewpoints found")

    # Extract the views' names
    for i in raw_views:
        viewname = helpers.xpath_fetch_unique(
            "./viewpoint", i[1], "viewpoint description"
        )
        assert viewname is not None
        viewname = C.RE_VIEWPOINT.search(viewname.attrib["href"])
        if viewname is not None:
            viewname = viewname.group(1)
        views.append((*i, urllib.parse.unquote(viewname or "")))

    if not views:
        raise ValueError("No viewpoints found")

    descriptors: list[tuple[pathlib.PurePosixPath, etree._Element, str]] = []
    for view in views:
        descriptors += [
            (view[0], d, view[2]) for d in C.XP_DESCRIPTORS(view[1])
        ]

    if len(descriptors) == 0:
        raise ValueError("Invalid XML: No diagrams found")

    for descriptor in descriptors:
        name = uid = None
        try:
            name = descriptor[1].attrib["name"]
            uid = descriptor[1].attrib["repPath"]
            if not uid.startswith("#"):
                raise ValueError("Malformed diagram reference: {uid!r}")

            diag_root = model[uid]
            if diag_root.tag not in DIAGRAM_ROOTS:
                continue

            # Extract styleclass from diagram description
            styledescription = helpers.xpath_fetch_unique(
                "./description[@href]", descriptor[1], "style description"
            )
            assert styledescription is not None
            styledescription = styledescription.attrib["href"]
            # This style description is something resembling XPath, e.g.
            #     platform:/[...]#//[...]/@ownedRepresentations[name='$$$']
            # The thing we're interested in is denoted as $$$ above.
            styleclass_match = C.RE_STYLECLASS.search(styledescription)
            if styleclass_match is None:
                styleclass: str | None = None
            else:
                styleclass = urllib.parse.unquote(styleclass_match.group(1))

            target_anchors = list(descriptor[1].iterchildren("target"))
            if len(target_anchors) != 1:
                raise RuntimeError(
                    f"Expected 1 <target> anchor, found {len(target_anchors)}"
                )
            target_href = target_anchors[0].get("href")
            if not target_href:
                raise RuntimeError("<target> anchor has no href")
            target = model.follow_link(descriptor[1], target_href)

            yield DiagramDescriptor(
                fragment=descriptor[0],
                name=name,
                styleclass=styleclass,
                uid=uid[1:],
                viewpoint=descriptor[2],
                target=target,
                descriptor=descriptor[1],
            )
        except Exception as err:
            C.LOGGER.warning(
                "Ignoring invalid diagram %s (%r): %s", uid, name, err
            )


def parse_diagrams(
    model: loader.MelodyLoader, **params: t.Any
) -> cabc.Iterator[diagram.Diagram]:
    """Parse all diagrams from the model."""
    for descriptor in enumerate_diagrams(model):
        C.LOGGER.info(
            "Extracting diagram %r of type %s",
            descriptor.name,
            descriptor.styleclass,
        )
        yield parse_diagram(model, descriptor, **params)


def parse_diagram(
    model: loader.MelodyLoader, descriptor: DiagramDescriptor, **params: t.Any
) -> diagram.Diagram:
    """Parse a single diagram from the model.

    Parameters
    ----------
    model
        A loaded model.
    descriptor
        A DiagramDescriptor as obtained from :func:`enumerate_diagrams`.
    """
    diag = diagram.Diagram(
        descriptor.name, styleclass=descriptor.styleclass, uuid=descriptor.uid
    )
    dgtree = model.follow_link(
        model.trees[descriptor.fragment].root, descriptor.uid
    )
    treedata = helpers.xpath_fetch_unique(
        C.XP_ANNOTATION_ENTRIES,
        dgtree,
        "ownedAnnotationsEntries with source GMF_DIAGRAMS",
        dgtree.attrib["uid"],
    )
    assert treedata is not None

    if treedata.attrib.get("measurementUnit", "Pixel") != "Pixel":
        C.LOGGER.warning(
            "Unknown measurement unit %r",
            treedata.attrib.get("measurementUnit"),
        )

    for data_elm in treedata.iterdescendants("children", "edges"):
        try:
            elm = _element_from_xml(
                C.ElementBuilder(
                    target_diagram=diag,
                    diagram_tree=dgtree,
                    data_element=data_elm,
                    melodyloader=model,
                    fragment=descriptor.fragment,
                )
            )
        except C.SkipObject:
            continue
        assert elm is not None
        diag.add_element(elm, False, force=True)

    if len(diag) == 0:
        C.LOGGER.error(
            "Deserialized diagram %r is empty", descriptor.name or diag.name
        )
    else:
        diag.calculate_viewport()
        _filters.applyfilters(
            _filters.FilterArguments(diag, dgtree, model, params)
        )
    return diag


def _element_from_xml(ebd: C.ElementBuilder) -> diagram.DiagramElement:
    """Construct a single diagram element from the model XML."""
    if ebd.data_element.get("element") is not None:
        factory = _semantic.from_xml
    else:
        factory = _visual.from_xml
    return factory(ebd)


if not t.TYPE_CHECKING:

    def __getattr__(key: str) -> t.Any:
        if key == "diagram":
            obj = diagram
        else:
            try:
                obj = getattr(diagram, key)
            except NameError:
                try:
                    obj = importlib.import_module(f"capellambse.diagram.{key}")
                except ImportError:
                    raise NameError(f"No name {key} in {__name__}") from None
        import warnings

        warnings.warn(
            (
                f"{__name__}.{key} is deprecated,"
                " import it from capellambse.diagram instead"
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return obj



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Various diagramming related tools.

This module is used to create diagrams, which can then be exported in
various formats (such as SVG).
"""
# isort: off
from ._vector2d import *

from .capstyle import *
from ._diagram import *
from ._json_enc import *



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0


# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

import capellambse

# -- Project information -----------------------------------------------------

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib
with open("../../pyproject.toml", "rb") as f:
    _metadata = tomllib.load(f)["project"]

project = "py-capellambse"
pypi = "capellambse"
author = _metadata["authors"][0]["name"]
copyright = f"{author} and the {_metadata['name']} contributors"
license = _metadata["license"]["text"]
install_requirements = _metadata["dependencies"]
python_requirement = _metadata["requires-python"]

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "capellambse.sphinx",
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_argparse_cli",
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
# exclude_patterns = []


# -- General information about the project -----------------------------------

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# The full version, including alpha/beta/rc tags.
version = capellambse.__version__
rst_epilog = f"""
.. |Project| replace:: {project}
.. |Version| replace:: {version}
"""


# -- Options for auto-doc ----------------------------------------------------
autoclass_content = "class"
autodoc_class_signature = "separated"
autodoc_typehints = "description"


# -- Options for napoleon ----------------------------------------------------
napoleon_custom_sections = ["Transaction arguments", "Well-known arguments"]
napoleon_google_docstring = False
napoleon_include_init_with_doc = True


# -- Options for Intersphinx output ------------------------------------------
intersphinx_mapping = {
    "lxml": ("https://lxml.de/apidoc/", None),
    "python": ("https://docs.python.org/3", None),
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "furo"
html_theme_options = {
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/DSD-DBS/py-capellambse",
            "html": '<img src="/_static/img/github-logo.svg"/>',
            "class": "",
        },
    ],
}
html_short_title = "py-capellambse"
html_show_sourcelink = False
html_context = {
    "dependencies": install_requirements,
    "py_req": python_requirement,
}

# -- Extra options for Furo theme --------------------------------------------

pygments_style = "tango"
pygments_dark_style = "monokai"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Options for CapellaMBSE-Sphinx ------------------------------------------
capellambse_model = "../capella-python-api/capella-python-api.aird"



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""The color palette and default style definitions used by Capella."""
from __future__ import annotations

__all__ = ["COLORS", "CSSdef", "STYLES", "RGB", "get_style"]

import logging
import typing as t

from capellambse import helpers

LOGGER = logging.getLogger(__name__)


class RGB(t.NamedTuple):
    """A color.

    Each color component (red, green, blue) is an integer in the range
    of 0..255 (inclusive).  The alpha channel is a float between 0.0 and
    1.0 (inclusive).  If it is 1, then the ``str()`` form does not
    include transparency information.
    """

    r: int = 0
    g: int = 0
    b: int = 0
    a: float = 1.0

    def __str__(self) -> str:
        return "#" + self.tohex()

    def tohex(self) -> str:
        # pylint: disable=unsubscriptable-object
        assert all(0 <= n <= 255 for n in self[:3])
        assert 0.0 <= self.a <= 1.0
        if self.a >= 1.0:
            return f"{self.r:02X}{self.g:02X}{self.b:02X}"
        return f"{self.r:02X}{self.g:02X}{self.b:02X}{int(self.a * 255):02X}"

    @classmethod
    def fromcss(cls, cssstring: str | RGB) -> RGB:
        """Create an RGB from a CSS color definition.

        Examples of recognized color definitions and their equivalent
        constructor calls::

            "rgb(10, 20, 30)" -> RGB(10, 20, 30)
            "rgba(50, 60, 70, 0.5)" -> RGB(50, 60, 70, 0.5)
            "#FF00FF" -> RGB(255, 0, 255)
            "#ff00ff" -> RGB(255, 0, 255)
            "#f0f" -> RGB(255, 0, 255)
            "#FF00FF80" -> RGB(255, 0, 255, 0.5)
            "#f0fa" -> RGB(255, 0, 255, 2/3)
        """
        if isinstance(cssstring, RGB):
            return cssstring

        cssstring = cssstring.strip().lower()
        if cssstring.startswith(("rgb(", "rgba(")) and cssstring.endswith(")"):
            return cls.fromcsv(cssstring[cssstring.find("(") : -1])
        if cssstring.startswith("#"):
            return cls.fromhex(cssstring[1:])
        raise ValueError(f"Bad CSS color: {cssstring!r}")

    @classmethod
    def fromcsv(cls, csvstring: str) -> RGB:
        """Create an RGB from a ``"r, g, b[, a]"`` string."""
        split = csvstring.split(",")
        if len(split) == 4:
            alpha = float(split.pop())
        else:
            alpha = 1.0
        if len(split) == 3:
            r, g, b = (int(c) for c in split)
            return cls(r, g, b, alpha)
        raise ValueError(f"Expected 3 or 4 values: {csvstring}")

    @classmethod
    def fromhex(cls, hexstring: str) -> RGB:
        """Create an RGB from a hexadecimal string.

        The string can have 3, 4, 6 or 8 hexadecimal characters.  In the
        cases of 3 and 6 characters, the alpha channel is set to 1.0
        (fully opaque) and the remaining characters are interpreted as
        the red, green and blue components.
        """
        if hexstring.startswith("#"):
            hs = hexstring[1:]
        else:
            hs = hexstring
        alpha = 1.0
        slen = len(hs)

        if slen == 4:
            hs, alpha = hs[:3], int(hs[3:], base=16) / 16
            slen = 3
        if slen == 3:
            r, g, b = (int(x * 2, base=16) for x in hs)
            return cls(r, g, b, alpha)

        if slen == 8:
            hs, alpha = hs[:6], int(hs[6:], base=16) / 255
            slen = 6
        if slen == 6:
            r, g, b = (
                int("".join(x), base=16) for x in helpers.ntuples(2, hs)
            )
            return cls(r, g, b, alpha)

        raise ValueError(
            "Invalid length of hex string, expected 3, 4, 6 or 8 characters"
        )


def get_style(diagramclass: str | None, objectclass: str) -> dict[str, t.Any]:
    r"""Fetch the default style for the given drawtype and styleclass.

    The style is returned as a dict with key-value pairs as used by CSS
    inside SVG graphics.

    All values contained in this dict are either of type :class:`str`,
    or of a class whose ``str()`` representation results in a valid CSS
    value for its respective key -- with one exception: color gradients.

    Flat colors are represented using the :class:`RGB` tuple subclass.
    Gradients are returned as a two-element list of :class:`RGB`\ s the
    first one is the color at the top of the object, the second one at
    the bottom.

    Parameters
    ----------
    diagramclass
        The style class of the diagram.
    objectclass
        A packed :class:`str` describing the element's type and style
        class in the form::

            Type.StyleClass

        The type can be: ``Box``, ``Edge``.  The style class can be any
        known style class.
    """
    if "symbol" in objectclass.lower():
        return {}

    if (
        objectclass not in STYLES["__GLOBAL__"]
        and diagramclass
        and objectclass not in STYLES.get(diagramclass, {})
    ):
        LOGGER.debug(
            "No default style for %r in %r", objectclass, diagramclass
        )
    try:
        retval = STYLES["__GLOBAL__"][objectclass].copy()
    except KeyError:
        retval = {}

    if diagramclass:
        retval.update(STYLES.get(diagramclass, {}).get(objectclass, {}))
    return retval


#: This dict maps the color names used by Capella to RGB tuples.
COLORS: dict[str, RGB] = {
    # System palette
    "black": RGB(0, 0, 0),
    "dark_gray": RGB(69, 69, 69),
    "dark_orange": RGB(224, 133, 3),
    "dark_purple": RGB(114, 73, 110),
    "gray": RGB(136, 136, 136),
    "light_purple": RGB(217, 196, 215),
    "light_yellow": RGB(255, 245, 181),
    "red": RGB(239, 41, 41),
    "white": RGB(255, 255, 255),
    # "Migration Palette" from common.odesign
    "_CAP_Activity_Border_Orange": RGB(91, 64, 64),
    "_CAP_Activity_Orange": RGB(247, 218, 116),
    "_CAP_Activity_Orange_min": RGB(255, 255, 197),
    "_CAP_Actor_Blue": RGB(198, 230, 255),
    "_CAP_Actor_Blue_label": RGB(0, 0, 0),
    "_CAP_Actor_Blue_min": RGB(218, 253, 255),
    "_CAP_Actor_Border_Blue": RGB(74, 74, 151),
    "_CAP_Association_Color": RGB(0, 0, 0),
    "_CAP_ChoicePseudoState_Border_Gray": RGB(0, 0, 0),
    "_CAP_ChoicePseudoState_Color": RGB(168, 168, 168),
    "_CAP_Class_Border_Brown": RGB(123, 105, 79),
    "_CAP_Class_Brown": RGB(232, 224, 210),
    "_CAP_CombinedFragment_Gray": RGB(242, 242, 242),
    "_CAP_Component_Blue": RGB(150, 177, 218),
    "_CAP_Component_Blue_min": RGB(195, 230, 255),
    "_CAP_Component_Border_Blue": RGB(74, 74, 151),
    "_CAP_Component_Label_Blue": RGB(74, 74, 151),
    "_CAP_ConfigurationItem_Gray": RGB(242, 238, 225),
    "_CAP_ConfigurationItem_Gray_min": RGB(249, 248, 245),
    "_CAP_Datatype_Border_Gray": RGB(103, 103, 103),
    "_CAP_Datatype_Gray": RGB(225, 223, 215),
    "_CAP_Datatype_LightBrown": RGB(232, 224, 210),
    "_CAP_Entity_Gray": RGB(221, 221, 200),
    "_CAP_Entity_Gray_border": RGB(69, 69, 69),
    "_CAP_Entity_Gray_label": RGB(0, 0, 0),
    "_CAP_Entity_Gray_min": RGB(249, 248, 245),
    "_CAP_ExchangeItem_Pinkkish": RGB(246, 235, 235),
    "_CAP_FCD": RGB(233, 243, 222),
    "_CAP_FCinFCD_Green": RGB(148, 199, 97),
    "_CAP_InterfaceDataPackage_LightGray": RGB(250, 250, 250),
    "_CAP_Interface_Border_Reddish": RGB(124, 61, 61),
    "_CAP_Interface_Pink": RGB(240, 221, 221),
    "_CAP_Lifeline_Gray": RGB(128, 128, 128),
    "_CAP_MSM_Mode_Gray": RGB(195, 208, 208),
    "_CAP_MSM_Mode_Gray_min": RGB(234, 239, 239),
    "_CAP_MSM_State_Gray": RGB(208, 208, 208),
    "_CAP_MSM_State_Gray_min": RGB(239, 239, 239),
    "_CAP_Mode_Gray": RGB(165, 182, 180),
    "_CAP_Node_Yellow": RGB(255, 252, 183),
    "_CAP_Node_Yellow_Border": RGB(123, 105, 79),
    "_CAP_Node_Yellow_Label": RGB(0, 0, 0),
    "_CAP_Node_Yellow_min": RGB(255, 255, 220),
    "_CAP_OperationalRole_Purple": RGB(203, 174, 200),
    "_CAP_Operational_Process_Reference_Orange": RGB(250, 239, 203),
    "_CAP_PhysicalPort_Yellow": RGB(255, 244, 119),
    "_CAP_StateMode_Border_Gray": RGB(117, 117, 117),
    "_CAP_StateTransition_Color": RGB(0, 0, 0),
    "_CAP_State_Gray": RGB(228, 228, 228),
    "_CAP_Unit_LightBrown": RGB(214, 197, 171),
    "_CAP_Value_LightBrown": RGB(254, 253, 250),
    "_CAP_xAB_Activity_Label_Orange": RGB(91, 64, 64),
    "_CAP_xAB_Function_Border_Green": RGB(9, 92, 46),
    "_CAP_xAB_Function_Green": RGB(197, 255, 166),
    "_CAP_xAB_Function_Label_Green": RGB(9, 92, 46),
    "_CAP_xBD_ControlNode": RGB(223, 223, 223),
    "_CAP_xDFB_Function_Border_Green": RGB(77, 137, 20),
    "_CAP_xDFB_Function_Green": RGB(197, 255, 166),
    "_CAP_xDFB_Function_Green_Label": RGB(0, 0, 0),
    "_CAP_xDFB_Function_Green_min": RGB(244, 255, 224),
    "_CAP_xDF_Activity_Label_Orange": RGB(0, 0, 0),
}


#: This dict contains the default styles that Capella applies, grouped
#: by the diagram class they belong to.
#:
#: The first level of keys are the diagrams' styleclasses.  The special
#: key "__GLOBAL__" applies to all diagrams.
#:
#: The second level contains the style definitions for each element that
#: can appear in the diagram.  The keys work in the following way:
#:
#:     Type.Class
#:
#: * ``Type`` is the element type; one of "Box" or "Edge" (note casing!)
#: * ``Class`` is the element's styleclass, e.g. "LogicalComponent"
#:
#: The ``Class`` and the preceding dot may be absent, in which case that
#: styling applies to all elements of that ``Type`` regardless of their
#: style class.
#:
#: The order of precedence for the four possible cases is the following,
#: from most to least important:
#:
#: 1. Diagram class specific, element type and class
#: 2. __GLOBAL__, element type and class
#: 3. Diagram class specific, only element type
#: 4. __GLOBAL__, only element type
CSSdef = t.Union[int, str, RGB, t.List[RGB], None]
STYLES: dict[str, dict[str, dict[str, CSSdef]]] = {
    "__GLOBAL__": {  # Global defaults
        "Box": {
            "fill": "transparent",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
        "Box.Annotation": {
            "fill": None,
            "stroke": None,
        },
        "Box.Constraint": {  # DT_Contraint [sic]
            "fill": COLORS["light_yellow"],
            "stroke": COLORS["gray"],
            "text_fill": COLORS["black"],
        },
        "Box.Note": {
            "fill": RGB(255, 255, 203),
            "stroke": RGB(255, 204, 102),
            "text_fill": RGB(0, 0, 0),
        },
        "Box.Requirement": {  # ReqVP_Requirement
            "fill": COLORS["light_purple"],
            "stroke": COLORS["dark_purple"],
            "text_fill": COLORS["black"],
        },
        "Box.Text": {
            "stroke": "transparent",
        },
        "Edge": {
            "stroke-width": 1,
        },
        "Edge.Connector": {
            "stroke": RGB(176, 176, 176),
            "stroke-dasharray": "1",
        },
        "Edge.Constraint": {
            "stroke": COLORS["black"],
            "stroke-dasharray": "1, 3",
            "marker-end": "FineArrowMark",
            "stroke-linecap": "round",
        },
        "Edge.Note": {
            "stroke": COLORS["black"],
            "stroke-dasharray": "1, 3",
        },
        "Edge.RequirementRelation": {  # ReqVP_IncomingRelation
            "stroke": COLORS["dark_purple"],
            "stroke-width": 2,
            "marker-end": "FineArrowMark",
            "stroke-dasharray": "5",
            "text_fill": COLORS["dark_purple"],
        },
    },
    "Class Diagram Blank": {
        "Box.Class": {  # DT_Class
            "fill": [COLORS["white"], COLORS["_CAP_Class_Brown"]],
            "rx": "10px",
            "ry": "10px",
            "stroke": COLORS["_CAP_Class_Border_Brown"],
            "text_fill": COLORS["black"],
        },
        "Box.DataPkg": {  # DT_DataPkg
            "fill": [
                COLORS["white"],
                COLORS["_CAP_InterfaceDataPackage_LightGray"],
            ],
            "stroke": COLORS["dark_gray"],
            "text_fill": COLORS["black"],
        },
        "Box.Enumeration": {  # DT_DataType
            "fill": COLORS["_CAP_Class_Brown"],
            "stroke": COLORS["_CAP_Datatype_Border_Gray"],
            "text_fill": COLORS["black"],
        },
        "Box.ExchangeItem": {  # DT_ExchangeItem
            "fill": COLORS["_CAP_ExchangeItem_Pinkkish"],
            "stroke": COLORS["_CAP_Interface_Border_Reddish"],
            "text_fill": COLORS["black"],
        },
        "Edge.Association": {
            "stroke": COLORS["_CAP_Association_Color"],
            "marker-end": "FineArrowMark",
        },
        "Edge.Aggregation": {
            "stroke": COLORS["_CAP_Association_Color"],
            "marker-start": "DiamondMark",
        },
        "Edge.Composition": {
            "stroke": COLORS["_CAP_Association_Color"],
            "marker-start": "FilledDiamondMark",
        },
        "Edge.ExchangeItemElement": {  # DT_ExchangeItemElement
            "stroke": COLORS["black"],
            "stroke-dasharray": "5",
            "text_fill": COLORS["black"],
            "marker-start": "DiamondMark",
            "marker-end": "ArrowMark",
        },
        "Edge.Generalization": {
            "stroke": COLORS["black"],
            "marker-end": "GeneralizationMark",
        },
    },
    "Contextual Capability": {
        "Edge.AbstractCapabilityExtend": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
        "Edge.AbstractCapabilityGeneralization": {
            "marker-end": "GeneralizationMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
        "Edge.AbstractCapabilityInclude": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
        "Edge.CapabilityExploitation": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
        "Edge.CapabilityInvolvement": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
        "Edge.MissionInvolvement": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
    },
    "Error": {},
    "Functional Chain Description": {
        "Box.Function": {
            "fill": COLORS["_CAP_xAB_Function_Green"],
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
        },
        "Edge.FunctionalExchange": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
        },
        "Edge.SequenceLink": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-dasharray": "5",
        },
    },
    "Logical Architecture Blank": {  # (from logical.odesign)
        **dict.fromkeys(
            ["Box.CP_IN", "Box.CP_OUT", "Box.CP_INOUT"],
            {
                "fill": COLORS["white"],
                "stroke": COLORS["black"],
            },
        ),
        "Box.CP_UNSET": {
            "fill": [COLORS["red"], COLORS["white"]],
        },
        "Box.FIP": {
            "fill": COLORS["dark_orange"],
            "stroke-width": 0,
        },
        "Box.FOP": {
            "fill": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 0,
        },
        "Box.LogicalActor": {  # Logical Actors
            "fill": [COLORS["_CAP_Actor_Blue_min"], COLORS["_CAP_Actor_Blue"]],
            "stroke": COLORS["_CAP_Actor_Border_Blue"],
            "text_fill": COLORS["_CAP_Actor_Blue_label"],
        },
        "Box.LogicalComponent": {  # LAB Logical Component
            "fill": [
                COLORS["_CAP_Component_Blue_min"],
                COLORS["_CAP_Component_Blue"],
            ],
            "stroke": COLORS["_CAP_Component_Border_Blue"],
            "text_fill": COLORS["_CAP_Component_Label_Blue"],
        },
        "Box.LogicalHumanActor": {  # LAB Logical Human Actor
            "fill": [
                COLORS["_CAP_Component_Blue_min"],
                COLORS["_CAP_Component_Blue"],
            ],
            "stroke": COLORS["_CAP_Component_Border_Blue"],
            "text_fill": COLORS["_CAP_Component_Label_Blue"],
        },
        "Box.LogicalHumanComponent": {  # LAB Logical Human Component
            "fill": [
                COLORS["_CAP_Component_Blue_min"],
                COLORS["_CAP_Component_Blue"],
            ],
            "stroke": COLORS["_CAP_Component_Border_Blue"],
            "text_fill": COLORS["_CAP_Component_Label_Blue"],
        },
        "Box.LogicalFunction": {  # LAB Logical Function
            "fill": COLORS["_CAP_xAB_Function_Green"],
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
            "text_fill": COLORS["_CAP_xAB_Function_Label_Green"],
        },
        "Edge.FunctionalExchange": {  # LAB DataFlow between Function
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 2,
            "text_fill": COLORS["_CAP_xAB_Function_Border_Green"],
        },
        "Edge.ComponentExchange": {  # LAB DataFlow between Logical Components
            "stroke": COLORS["_CAP_Component_Border_Blue"],
            "stroke-width": 2,
            "text_fill": COLORS["_CAP_Component_Border_Blue"],
        },
        "Edge.FIPAllocation": {  # LAB PortAllocation
            "stroke": COLORS["dark_orange"],
            "stroke-width": 2,
            "stroke-dasharray": "5",
        },
        "Edge.FOPAllocation": {
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 2,
            "stroke-dasharray": "5",
        },
    },
    "Logical Data Flow Blank": {
        "Box.FIP": {
            "fill": COLORS["dark_orange"],
            "stroke-width": 0,
        },
        "Box.FOP": {
            "fill": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 0,
        },
        "Box.LogicalFunction": {  # LDFB_Function
            "fill": [
                COLORS["_CAP_xDFB_Function_Green_min"],
                COLORS["_CAP_xDFB_Function_Green"],
            ],
            "rx": "10px",
            "ry": "10px",
            "stroke": COLORS["_CAP_xDFB_Function_Border_Green"],
            "text_fill": COLORS["_CAP_xDFB_Function_Green_Label"],
        },
        "Edge.FunctionalExchange": {  # LDFB_Exchange
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 2,
            "text_fill": COLORS["_CAP_xAB_Function_Border_Green"],
        },
    },
    "Missions Capabilities Blank": {
        "Box.SystemComponent": {
            "fill": [COLORS["_CAP_Actor_Blue_min"], COLORS["_CAP_Actor_Blue"]],
            "stroke": COLORS["_CAP_Actor_Border_Blue"],
            "text_fill": COLORS["_CAP_Actor_Blue_label"],
        },
        "Box.SystemActor": {
            "fill": [COLORS["_CAP_Actor_Blue_min"], COLORS["_CAP_Actor_Blue"]],
            "stroke": COLORS["_CAP_Actor_Border_Blue"],
            "text_fill": COLORS["_CAP_Actor_Blue_label"],
        },
        "Box.SystemHumanActor": {
            "fill": [COLORS["_CAP_Actor_Blue_min"], COLORS["_CAP_Actor_Blue"]],
            "stroke": COLORS["_CAP_Actor_Border_Blue"],
            "text_fill": COLORS["_CAP_Actor_Blue_label"],
        },
        "Edge.AbstractCapabilityExtend": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
        "Edge.AbstractCapabilityGeneralization": {
            "marker-end": "GeneralizationMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
        "Edge.AbstractCapabilityInclude": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
        "Edge.CapabilityExploitation": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
        "Edge.CapabilityInvolvement": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
        "Edge.MissionInvolvement": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-width": 1,
        },
    },
    "Mode State Machine": {  # (from common.odesign)
        "Box.ChoicePseudoState": {
            "fill": COLORS["_CAP_ChoicePseudoState_Color"],
            "stroke": COLORS["_CAP_ChoicePseudoState_Border_Gray"],
        },
        "Box.ForkPseudoState": {
            "fill": COLORS["black"],
        },
        "Box.JoinPseudoState": {
            "fill": COLORS["black"],
        },
        "Box.Mode": {
            "fill": COLORS["_CAP_MSM_Mode_Gray_min"],
            "stroke": COLORS["dark_gray"],
        },
        "Box.ModeRegion": {
            "fill": [
                COLORS["_CAP_MSM_Mode_Gray_min"],
                COLORS["_CAP_MSM_Mode_Gray"],
            ],
            "stroke": COLORS["dark_gray"],
        },
        "Box.State": {
            "fill": COLORS["_CAP_MSM_State_Gray_min"],
            "stroke": COLORS["dark_gray"],
        },
        "Box.StateRegion": {
            "fill": [
                COLORS["_CAP_MSM_State_Gray_min"],
                COLORS["_CAP_MSM_State_Gray"],
            ],
            "stroke": COLORS["dark_gray"],
        },
        "Edge.StateTransition": {
            "marker-end": "FineArrowMark",
        },
    },
    "Operational Capabilities Blank": {
        "Box.Entity": {
            "fill": [
                COLORS["_CAP_Entity_Gray_min"],
                COLORS["_CAP_Entity_Gray"],
            ],
            "stroke": COLORS["_CAP_Entity_Gray_border"],
            "text_fill": COLORS["_CAP_Entity_Gray_label"],
        },
        "Box.OperationalActor": {
            "fill": [
                COLORS["_CAP_Entity_Gray_min"],
                COLORS["_CAP_Entity_Gray"],
            ],
            "stroke": COLORS["_CAP_Entity_Gray_border"],
            "text_fill": COLORS["_CAP_Entity_Gray_label"],
        },
        "Edge.AbstractCapabilityExtend": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
        },
        "Edge.AbstractCapabilityGeneralization": {
            "marker-end": "GeneralizationMark",
            "stroke": COLORS["black"],
        },
        "Edge.AbstractCapabilityInclude": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
        },
        "Edge.CommunicationMean": {
            "marker-end": "ArrowMark",
            "stroke": COLORS["gray"],
        },
        "Edge.Entity": {},
        "Edge.EntityOperationalCapabilityInvolvement": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
        },
        "Edge.OperationalActor": {},
    },
    "Operational Entity Blank": {  # "Operational Architecture Blank" in GUI
        "Box.Entity": {
            "fill": [
                COLORS["_CAP_Entity_Gray_min"],
                COLORS["_CAP_Entity_Gray"],
            ],
            "stroke": COLORS["_CAP_Entity_Gray_border"],
            "text_fill": COLORS["_CAP_Entity_Gray_label"],
        },
        "Box.OperationalActivity": {
            "fill": COLORS["_CAP_Activity_Orange"],
            "stroke": COLORS["_CAP_Activity_Border_Orange"],
            "text_fill": COLORS["_CAP_xAB_Activity_Label_Orange"],
        },
        "Box.OperationalActor": {
            "fill": [
                COLORS["_CAP_Entity_Gray_min"],
                COLORS["_CAP_Entity_Gray"],
            ],
            "stroke": COLORS["_CAP_Entity_Gray_border"],
            "text_fill": COLORS["_CAP_Entity_Gray_label"],
        },
        "Box.Role": {
            "fill": [COLORS["white"], COLORS["_CAP_OperationalRole_Purple"]],
            "stroke": COLORS["dark_purple"],
            "text_fill": COLORS["black"],
        },
        "Edge.CommunicationMean": {
            "marker-end": "ArrowMark",
            "stroke": COLORS["dark_gray"],
        },
        "Edge.FunctionalExchange": {
            "marker-end": "ArrowMark",
            "stroke": COLORS["_CAP_Activity_Border_Orange"],
        },
    },
    "Operational Entity Breakdown": {
        "Box.Entity": {
            "fill": COLORS["_CAP_Entity_Gray"],
            "stroke": COLORS["_CAP_Entity_Gray_border"],
            "text_fill": COLORS["_CAP_Entity_Gray_label"],
        },
        "Edge.Entity": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["gray"],
        },
        "Edge.OperationalActor": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["gray"],
        },
    },
    "Operational Process Description": {
        "Box.OperationalActivity": {
            "fill": COLORS["_CAP_Activity_Orange"],
            "stroke": COLORS["_CAP_Activity_Border_Orange"],
        },
        "Edge.OperationalExchange": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["_CAP_Activity_Border_Orange"],
        },
        "Edge.SequenceLink": {
            "marker-end": "FineArrowMark",
            "stroke": COLORS["black"],
            "stroke-dasharray": "5",
        },
    },
    "Operational Activity Interaction Blank": {
        "Box.OperationalActivity": {
            "fill": COLORS["_CAP_Activity_Orange"],
            "rx": "10px",
            "ry": "10px",
            "stroke": COLORS["_CAP_Activity_Border_Orange"],
            "text_fill": COLORS["_CAP_xAB_Activity_Label_Orange"],
        },
        "Box.OperationalProcess": {
            "stroke": COLORS["black"],
            "text_fill": COLORS["black"],
        },
        "Edge.FunctionalExchange": {
            "marker-end": "ArrowMark",
            "stroke-width": 2,
            "stroke": COLORS["_CAP_Activity_Border_Orange"],
        },
    },
    "Physical Architecture Blank": {
        **dict.fromkeys(
            ["Box.CP_IN", "Box.CP_OUT", "Box.CP_INOUT"],
            {
                "fill": COLORS["white"],
                "stroke": COLORS["black"],
            },
        ),
        "Box.FIP": {
            "fill": COLORS["dark_orange"],
            "stroke-width": 0,
        },
        "Box.FOP": {
            "fill": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 0,
        },
        "Box.PP": {
            "fill": COLORS["_CAP_PhysicalPort_Yellow"],
            "stroke": COLORS["_CAP_Class_Border_Brown"],
        },
        "Box.PhysicalComponent": {
            "fill": [
                COLORS["_CAP_Node_Yellow_min"],
                COLORS["_CAP_Node_Yellow"],
            ],
            "stroke": COLORS["_CAP_Node_Yellow_Border"],
            "text_fill": COLORS["_CAP_Node_Yellow_Label"],
        },
        "Box.PhysicalFunction": {
            "fill": COLORS["_CAP_xAB_Function_Green"],
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
        },
        "Edge.ComponentExchange": {
            "stroke": COLORS["_CAP_Component_Border_Blue"],
            "stroke-width": 2,
            "text_fill": COLORS["_CAP_Component_Border_Blue"],
        },
        "Edge.FunctionalExchange": {
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 2,
            "text_fill": COLORS["_CAP_xAB_Function_Border_Green"],
        },
        "Edge.PhysicalLink": {
            "stroke": COLORS["red"],
            "stroke-width": 2,
            "text_fill": COLORS["red"],
        },
    },
    "Physical Data Flow Blank": {
        "Box.FIP": {
            "fill": COLORS["dark_orange"],
            "stroke-width": 0,
        },
        "Box.FOP": {
            "fill": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 0,
        },
        "Box.PhysicalFunction": {
            "fill": COLORS["_CAP_xAB_Function_Green"],
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
        },
        "Edge.FunctionalExchange": {
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 2,
            "text_fill": COLORS["_CAP_xAB_Function_Border_Green"],
        },
    },
    "System Architecture Blank": {
        **dict.fromkeys(
            ["Box.CP_IN", "Box.CP_OUT", "Box.CP_INOUT"],
            {
                "fill": COLORS["white"],
                "stroke": COLORS["black"],
            },
        ),
        "Box.FIP": {
            "fill": COLORS["dark_orange"],
            "stroke-width": 0,
        },
        "Box.FOP": {
            "fill": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 0,
        },
        "Box.SystemActor": {
            "fill": [COLORS["_CAP_Actor_Blue_min"], COLORS["_CAP_Actor_Blue"]],
            "stroke": COLORS["_CAP_Actor_Border_Blue"],
            "text_fill": COLORS["_CAP_Actor_Blue_label"],
        },
        "Box.SystemComponent": {
            "fill": [COLORS["_CAP_Actor_Blue_min"], COLORS["_CAP_Actor_Blue"]],
            "stroke": COLORS["_CAP_Actor_Border_Blue"],
            "text_fill": COLORS["_CAP_Actor_Blue_label"],
        },
        "Box.SystemHumanActor": {
            "fill": [COLORS["_CAP_Actor_Blue_min"], COLORS["_CAP_Actor_Blue"]],
            "stroke": COLORS["_CAP_Actor_Border_Blue"],
            "text_fill": COLORS["_CAP_Actor_Blue_label"],
        },
        "Box.SystemFunction": {
            "fill": COLORS["_CAP_xAB_Function_Green"],
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
        },
        "Edge.ComponentExchange": {
            "stroke": COLORS["_CAP_Component_Border_Blue"],
            "stroke-width": 2,
            "text_fill": COLORS["_CAP_Component_Border_Blue"],
        },
        "Edge.FunctionalExchange": {
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 2,
            "text_fill": COLORS["_CAP_xAB_Function_Border_Green"],
        },
        "Edge.FIPAllocation": {
            "stroke": COLORS["dark_orange"],
            "stroke-width": 2,
            "stroke-dasharray": "5",
        },
        "Edge.FOPAllocation": {
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 2,
            "stroke-dasharray": "5",
        },
    },
    "System Data Flow Blank": {
        "Box.FIP": {
            "fill": COLORS["dark_orange"],
            "stroke-width": 0,
        },
        "Box.FOP": {
            "fill": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 0,
        },
        "Box.SystemFunction": {  # SDFB_Function
            "fill": [
                COLORS["_CAP_xDFB_Function_Green_min"],
                COLORS["_CAP_xDFB_Function_Green"],
            ],
            "rx": "10px",
            "ry": "10px",
            "stroke": COLORS["_CAP_xDFB_Function_Border_Green"],
            "text_fill": COLORS["_CAP_xDFB_Function_Green_Label"],
        },
        "Edge.FunctionalExchange": {  # SDFB_Exchange
            "stroke": COLORS["_CAP_xAB_Function_Border_Green"],
            "stroke-width": 2,
            "text_fill": COLORS["_CAP_xAB_Function_Border_Green"],
        },
    },
}



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

import json
import pathlib
import sys

import pytest

import capellambse
from capellambse import aird, diagram, loader


class TestAIRDBasicFunctionality:
    test_model = (
        pathlib.Path(__file__).parent / "data" / "parser" / "TestItems.aird"
    )
    test_diagram = "[LAB] Logical System"
    test_json = test_model.with_suffix(".json")
    test_repr = test_model.with_suffix(".repr.txt")
    test_txt = test_model.with_suffix(".txt")

    @pytest.fixture
    def model(self):
        return loader.MelodyLoader(self.test_model)

    @pytest.fixture
    def diagram_under_test(self, model):
        descriptor = next(aird.enumerate_diagrams(model), None)
        assert descriptor is not None
        yield aird.parse_diagram(model, descriptor)

    def test_parsing_all_diagrams_does_not_raise_exceptions(
        self, model, caplog
    ):
        del caplog
        i = 0
        for i, _ in enumerate(aird.parse_diagrams(model), start=1):
            pass
        assert i == 1

    @pytest.mark.xfail(
        sys.platform not in {"win32", "cygwin"},
        reason="Expected rendering inaccuracies on non-Windows platforms",
    )
    @pytest.mark.skip(reason="Currently broken")
    def test_json_output_matches_expected_output(
        self, diagram_under_test, caplog
    ):
        del caplog
        expected = self.test_json.read_text().strip()
        encoder = diagram.DiagramJSONEncoder(indent=4)

        actual = encoder.encode(diagram_under_test)

        assert actual == expected

    @pytest.mark.xfail(
        sys.platform not in {"win32", "cygwin"},
        reason="Expected rendering inaccuracies on non-Windows platforms",
    )
    @pytest.mark.skip(reason="Currently broken")
    def test_plain_text_representation_matches_expected_output(
        self, diagram_under_test, caplog
    ):
        del caplog
        expected = self.test_txt.read_text()
        actual = str(diagram_under_test)
        assert actual + "\n" == expected

    @pytest.mark.xfail(
        sys.platform not in {"win32", "cygwin"},
        reason="Expected rendering inaccuracies on non-Windows platforms",
    )
    @pytest.mark.skip(reason="Currently broken")
    def test_python_code_representation_matches_expected_output(
        self, diagram_under_test, caplog
    ):
        del caplog
        expected = self.test_repr.read_text().strip()
        actual = repr(diagram_under_test)
        assert actual == expected


def test_airdparser_msm_produces_valid_json_without_error(
    model: capellambse.MelodyModel,
):
    diag_name = "[MSM] States of Functional Human Being"
    all_diagrams = aird.enumerate_diagrams(model._loader)
    descriptor = next(i for i in all_diagrams if i.name == diag_name)
    parsed = aird.parse_diagram(model._loader, descriptor)

    generated_json = diagram.DiagramJSONEncoder(indent=4).encode(parsed)
    json.loads(generated_json)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

__all__ = [
    "AbstractRequirementsAttribute",
    "AbstractRequirementsRelation",
    "AbstractType",
    "AttributeDefinition",
    "AttributeDefinitionEnumeration",
    "BooleanValueAttribute",
    "DataTypeDefinition",
    "DateValueAttribute",
    "EnumDataTypeDefinition",
    "EnumValue",
    "EnumerationValueAttribute",
    "IntegerValueAttribute",
    "ModuleType",
    "RealValueAttribute",
    "RelationType",
    "RelationsList",
    "ReqIFElement",
    "Requirement",
    "RequirementType",
    "RequirementsFolder",
    "RequirementsIncRelation",
    "RequirementsIntRelation",
    "RequirementsModule",
    "RequirementsOutRelation",
    "RequirementsTypesFolder",
    "StringValueAttribute",
    "XT_FOLDER",
    "XT_INC_RELATION",
    "XT_INT_RELATION",
    "XT_MODULE",
    "XT_MODULE_TYPE",
    "XT_OUT_RELATION",
    "XT_RELATION_TYPE",
    "XT_REQUIREMENT",
    "XT_REQ_ATTRIBUTES",
    "XT_REQ_ATTR_BOOLEANVALUE",
    "XT_REQ_ATTR_DATEVALUE",
    "XT_REQ_ATTR_ENUMVALUE",
    "XT_REQ_ATTR_INTEGERVALUE",
    "XT_REQ_ATTR_REALVALUE",
    "XT_REQ_ATTR_STRINGVALUE",
    "XT_REQ_TYPE",
    "XT_REQ_TYPES",
    "XT_REQ_TYPES_DATA_DEF",
    "XT_REQ_TYPES_F",
    "XT_REQ_TYPE_ATTR_DEF",
    "XT_REQ_TYPE_ATTR_ENUM",
    "XT_REQ_TYPE_ENUM",
    "XT_REQ_TYPE_ENUM_DEF",
    "init",
]

import collections.abc as cabc
import logging
import os
import re
import typing as t

import markupsafe
from lxml import etree

import capellambse.model
import capellambse.model.common as c
from capellambse.loader import xmltools
from capellambse.model import crosslayer

from . import exporter

XT_REQUIREMENT = "Requirements:Requirement"
XT_REQ_ATTR_STRINGVALUE = "Requirements:StringValueAttribute"
XT_REQ_ATTR_REALVALUE = "Requirements:RealValueAttribute"
XT_REQ_ATTR_INTEGERVALUE = "Requirements:IntegerValueAttribute"
XT_REQ_ATTR_DATEVALUE = "Requirements:DateValueAttribute"
XT_REQ_ATTR_BOOLEANVALUE = "Requirements:BooleanValueAttribute"
XT_REQ_ATTR_ENUMVALUE = "Requirements:EnumerationValueAttribute"
XT_REQ_ATTRIBUTES = {
    XT_REQ_ATTR_ENUMVALUE,
    XT_REQ_ATTR_STRINGVALUE,
    XT_REQ_ATTR_REALVALUE,
    XT_REQ_ATTR_INTEGERVALUE,
    XT_REQ_ATTR_DATEVALUE,
    XT_REQ_ATTR_BOOLEANVALUE,
}
XT_INC_RELATION = "CapellaRequirements:CapellaIncomingRelation"
XT_OUT_RELATION = "CapellaRequirements:CapellaOutgoingRelation"
XT_INT_RELATION = "Requirements:InternalRelation"
XT_MODULE = "CapellaRequirements:CapellaModule"
XT_FOLDER = "Requirements:Folder"

XT_REQ_TYPES_F = "CapellaRequirements:CapellaTypesFolder"
XT_REQ_TYPES_DATA_DEF = "Requirements:DataTypeDefinition"
XT_REQ_TYPE = "Requirements:RequirementType"
XT_RELATION_TYPE = "Requirements:RelationType"
XT_MODULE_TYPE = "Requirements:ModuleType"
XT_REQ_TYPE_ENUM = "Requirements:EnumerationDataTypeDefinition"
XT_REQ_TYPE_ATTR_ENUM = "Requirements:EnumValue"
XT_REQ_TYPE_ATTR_DEF = "Requirements:AttributeDefinition"
XT_REQ_TYPE_ENUM_DEF = "Requirements:AttributeDefinitionEnumeration"
XT_REQ_TYPES = {
    XT_REQ_TYPES_F,
    XT_REQ_TYPES_DATA_DEF,
    XT_REQ_TYPE,
    XT_RELATION_TYPE,
    XT_MODULE_TYPE,
    XT_REQ_TYPE_ENUM,
    XT_REQ_TYPE_ATTR_ENUM,
    XT_REQ_TYPE_ATTR_DEF,
    XT_REQ_TYPE_ENUM_DEF,
}

logger = logging.getLogger("reqif")


class RequirementsRelationAccessor(
    c.WritableAccessor["AbstractRequirementsRelation"]
):
    """Searches for requirement relations in the architecture layer."""

    # pylint: disable=abstract-method  # Only partially implemented for now

    __slots__ = ("aslist",)

    def __init__(self, *args, **kw) -> None:
        super().__init__(
            *args, **kw, aslist=c.ElementList, single_attr="long_name"
        )

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        rel_objs = list(
            obj._model._loader.iterchildren_xt(obj._element, XT_INC_RELATION)
        )

        for i in obj._model._loader.iterall_xt(XT_OUT_RELATION):
            if RequirementsOutRelation.from_model(obj._model, i).source == obj:
                rel_objs.append(i)

        for i in obj._model._loader.iterall_xt(XT_INT_RELATION):
            rel = RequirementsIntRelation.from_model(obj._model, i)
            if obj in (rel.source, rel.target):
                rel_objs.append(i)
        return self._make_list(obj, rel_objs)

    def _make_list(self, parent_obj, elements):
        assert self.aslist is not None
        return self.aslist(
            parent_obj._model, elements, c.GenericElement, parent=parent_obj
        )

    def create(
        self,
        elmlist: c.ElementListCouplingMixin,
        /,
        *type_hints: str | None,
        **kw: t.Any,
    ) -> RequirementsIntRelation | RequirementsIncRelation:
        if "target" not in kw:
            raise TypeError("No `target` for new requirement relation")
        cls, xtype = self._find_relation_type(kw["target"])
        parent = elmlist._parent._element
        with elmlist._model._loader.new_uuid(parent) as uuid:
            return cls(
                elmlist._model,
                parent,
                **kw,
                source=elmlist._parent,
                uuid=uuid,
                xtype=xtype,
            )

    def _find_relation_type(
        self, target: c.GenericElement
    ) -> tuple[type[RequirementsIntRelation | RequirementsIncRelation], str]:
        if isinstance(target, Requirement):
            return (
                RequirementsIntRelation,
                XT_INT_RELATION,
            )
        elif isinstance(target, ReqIFElement):
            raise TypeError(
                "Cannot create relations to targets of type"
                f" {type(target).__name__}"
            )
        else:
            return (
                RequirementsIncRelation,
                XT_INC_RELATION,
            )


class ElementRelationAccessor(
    c.WritableAccessor["AbstractRequirementsRelation"]
):
    """Provides access to RequirementsRelations of a GenericElement."""

    # pylint: disable=abstract-method  # Only partially implemented for now

    __slots__ = ("aslist",)

    def __init__(self) -> None:
        super().__init__(aslist=RelationsList, single_attr="long_name")

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        relations: list[etree._Element] = []
        for relation in obj._model.search(
            XT_INC_RELATION, XT_INT_RELATION, XT_OUT_RELATION
        ):
            if obj in (relation.source, relation.target):
                relations.append(relation._element)
        return self._make_list(obj, relations)

    def _make_list(self, parent_obj, elements):
        assert self.aslist is not None
        return self.aslist(
            parent_obj._model,
            elements,
            None,
            parent=parent_obj,
            source=parent_obj,
        )


class ReqIFElement(c.GenericElement):
    """Attributes shared by all ReqIF elements."""

    identifier = xmltools.AttributeProperty(
        "_element", "ReqIFIdentifier", optional=True
    )
    long_name = xmltools.AttributeProperty(
        "_element", "ReqIFLongName", optional=True
    )
    description: str = xmltools.AttributeProperty(  # type: ignore[assignment]
        "_element", "ReqIFDescription", optional=True
    )
    name = xmltools.AttributeProperty("_element", "ReqIFName", optional=True)
    prefix = xmltools.AttributeProperty(
        "_element", "ReqIFPrefix", optional=True
    )
    type: c.Accessor = property(lambda _: None)  # type: ignore[assignment]

    def _short_repr_(self) -> str:  # pragma: no cover
        mytype = type(self).__name__
        parent = self._element
        if self.xtype in XT_REQ_TYPES:
            return f'<{mytype} {parent.get("ReqIFLongName")!r} ({self.uuid})>'

        name = (
            parent.get("ReqIFName")
            or parent.get("ReqIFChapterName")
            or parent.get("ReqIFLongName")
            or ""
        )
        return f"<{mytype} {name!r} ({self.uuid})>"

    def _short_html_(self) -> markupsafe.Markup:
        name = markupsafe.Markup.escape(self.name or self.long_name)
        return markupsafe.Markup(self._wrap_short_html(f" &quot;{name}&quot;"))


@c.xtype_handler(None, XT_REQ_TYPES_DATA_DEF)
class DataTypeDefinition(ReqIFElement):
    """A data type definition for requirement types."""

    _xmltag = "ownedDefinitionTypes"


@c.xtype_handler(None, XT_REQ_TYPE_ATTR_DEF)
class AttributeDefinition(ReqIFElement):
    """An attribute definition for requirement types."""

    _xmltag = "ownedAttributes"

    data_type = c.AttrProxyAccessor(DataTypeDefinition, "definitionType")


class AbstractRequirementsAttribute(c.GenericElement):
    _xmltag = "ownedAttributes"

    definition = c.AttrProxyAccessor(AttributeDefinition, "definition")

    value: xmltools.AttributeProperty | c.AttrProxyAccessor

    def __repr__(self) -> str:
        return self._short_repr_()

    def _short_repr_(self) -> str:
        mytype = type(self).__name__
        if self.definition is not None:
            return f"<{mytype} {self.definition.long_name!r} ({self.uuid})>"
        default_name = re.sub(r"(?<!^)([A-Z])", r" \1", mytype)
        return f"<{mytype} [{default_name}] ({self.uuid})>"

    def _short_html_(self) -> markupsafe.Markup:
        if self.definition is not None:
            name = markupsafe.Markup.escape(self.definition.long_name)
        else:
            name = markupsafe.Markup.escape("None")
        return markupsafe.Markup(self._wrap_short_html(f" &quot;{name}&quot;"))


class AttributeAccessor(c.DirectProxyAccessor[AbstractRequirementsAttribute]):
    def __init__(self) -> None:
        super().__init__(
            c.GenericElement,  # type: ignore[arg-type]
            XT_REQ_ATTRIBUTES,
            aslist=c.MixedElementList,
            list_extra_args={
                "mapkey": "definition.long_name",
                "mapvalue": "value",
            },
        )

    def _match_xtype(  # type: ignore[override]
        self, type_: str
    ) -> tuple[type, str]:
        type_ = type_.lower()
        try:
            return _attr_type_hints[type_]
        except KeyError:
            raise ValueError(f"Invalid type hint given: {type_!r}") from None


class RelationsList(c.ElementList["AbstractRequirementsRelation"]):
    def __init__(
        self,
        model: capellambse.MelodyModel,
        elements: list[etree._Element],
        elemclass: type[t.Any] | None = None,
        *,
        source: c.ModelObject,
    ) -> None:
        del elemclass
        super().__init__(model, elements, AbstractRequirementsRelation)
        self._source = source

    @t.overload
    def __getitem__(self, idx: int | str) -> AbstractRequirementsRelation:
        ...

    @t.overload
    def __getitem__(
        self, idx: slice
    ) -> c.ElementList[AbstractRequirementsRelation]:
        ...

    def __getitem__(self, idx):
        rel = super().__getitem__(idx)
        if isinstance(rel, c.ElementList):
            return rel

        assert isinstance(rel, AbstractRequirementsRelation)
        assert self._source in (rel.source, rel.target)
        if self._source == rel.source:
            return rel.target
        else:
            return rel.source

    def by_relation_type(self, reltype: str) -> RelationsList:
        matches = []
        for elm in self._elements:
            rel_elm = c.GenericElement.from_model(self._model, elm)
            assert isinstance(rel_elm, AbstractRequirementsRelation)
            if rel_elm.type is not None and rel_elm.type.name == reltype:
                matches.append(elm)
        return self._newlist(matches)

    def by_relation_class(
        self, class_: t.Literal["incoming", "outgoing", "internal"]
    ) -> RelationsList:
        relation_types = {
            "incoming": RequirementsIncRelation,
            "outgoing": RequirementsOutRelation,
            "internal": RequirementsIntRelation,
        }
        matches: list[etree._Element] = []
        for elm in self._elements:
            rel_elm = c.GenericElement.from_model(self._model, elm)
            if isinstance(rel_elm, relation_types[class_]):
                matches.append(rel_elm._element)
        return self._newlist(matches)

    def _newlist(self, elements: list[etree._Element]) -> RelationsList:
        listtype = self._newlist_type()
        assert issubclass(listtype, RelationsList)
        return listtype(self._model, elements, source=self._source)


@c.xtype_handler(None, XT_REQ_ATTR_BOOLEANVALUE)
class BooleanValueAttribute(AbstractRequirementsAttribute):
    """A string value attribute."""

    value = xmltools.BooleanAttributeProperty("_element", "value")


@c.xtype_handler(None, XT_REQ_ATTR_DATEVALUE)
class DateValueAttribute(AbstractRequirementsAttribute):
    """A value attribute that stores a date and time."""

    value = xmltools.DatetimeAttributeProperty(
        "_element", "value", optional=True
    )


@c.xtype_handler(None, XT_REQ_ATTR_INTEGERVALUE)
class IntegerValueAttribute(AbstractRequirementsAttribute):
    """An integer value attribute."""

    value = xmltools.AttributeProperty(
        "_element", "value", returntype=int, default=0
    )


@c.xtype_handler(None, XT_REQ_ATTR_REALVALUE)
class RealValueAttribute(AbstractRequirementsAttribute):
    """A floating-point number value attribute."""

    value = xmltools.AttributeProperty(
        "_element", "value", returntype=float, default=0.0
    )


@c.xtype_handler(None, XT_REQ_ATTR_STRINGVALUE)
class StringValueAttribute(AbstractRequirementsAttribute):
    """A string value attribute."""

    value = xmltools.AttributeProperty("_element", "value", default="")


@c.xtype_handler(None, XT_REQ_TYPE_ATTR_ENUM)
@c.attr_equal("long_name")
class EnumValue(ReqIFElement):
    """An enumeration value for :class:`EnumDataTypeDefinition`."""

    _xmltag = "specifiedValues"

    def __str__(self) -> str:
        return self.long_name


@c.xtype_handler(None, XT_REQ_TYPE_ENUM)
class EnumDataTypeDefinition(ReqIFElement):
    """An enumeration data type definition for requirement types."""

    _xmltag = "ownedDefinitionTypes"

    values = c.DirectProxyAccessor(
        EnumValue,
        XT_REQ_TYPE_ATTR_ENUM,
        aslist=c.ElementList,
        single_attr="long_name",
    )


@c.xtype_handler(None, XT_REQ_TYPE_ENUM_DEF)
class AttributeDefinitionEnumeration(ReqIFElement):
    """An enumeration attribute definition for requirement types."""

    _xmltag = "ownedAttributes"

    data_type = c.AttrProxyAccessor(EnumDataTypeDefinition, "definitionType")
    multi_valued = xmltools.BooleanAttributeProperty(
        "_element",
        "multiValued",
        __doc__=(
            "Boolean flag for setting multiple enumeration values on"
            " the attribute"
        ),
    )


@c.xtype_handler(None, XT_REQ_ATTR_ENUMVALUE)
class EnumerationValueAttribute(AbstractRequirementsAttribute):
    """An enumeration attribute."""

    definition = c.AttrProxyAccessor(
        AttributeDefinitionEnumeration, "definition"  # type: ignore[arg-type]
    )
    values = c.AttrProxyAccessor(EnumValue, "values", aslist=c.ElementList)

    @property
    def value(self):
        vals = self.values
        if len(vals) > 1:
            raise TypeError("Multi-value enumeration, use `.values` instead")
        if len(vals) == 1:
            return vals[0]
        return None  # TODO return a default value?


@c.attr_equal("long_name")
class AbstractType(ReqIFElement):
    owner = c.ParentAccessor(c.GenericElement)
    attribute_definitions = c.DirectProxyAccessor(
        c.GenericElement,
        (XT_REQ_TYPE_ATTR_DEF, XT_REQ_TYPE_ENUM_DEF),
        aslist=c.MixedElementList,
    )


@c.xtype_handler(None, XT_MODULE_TYPE)
class ModuleType(AbstractType):
    """A requirement-module type."""

    _xmltag = "ownedTypes"


@c.xtype_handler(None, XT_RELATION_TYPE)
class RelationType(AbstractType):
    """A requirement-relation type."""

    _xmltag = "ownedTypes"


@c.xtype_handler(None, XT_REQ_TYPE)
class RequirementType(AbstractType):
    """A requirement type."""

    _xmltag = "ownedTypes"


@c.xtype_handler(None, XT_REQUIREMENT)
class Requirement(ReqIFElement):
    """A ReqIF Requirement."""

    _xmltag = "ownedRequirements"

    owner = c.ParentAccessor(c.GenericElement)

    chapter_name = xmltools.AttributeProperty(
        "_element", "ReqIFChapterName", optional=True
    )
    foreign_id = xmltools.AttributeProperty(
        "_element", "ReqIFForeignID", optional=True, returntype=int
    )
    text = xmltools.HTMLAttributeProperty(
        "_element", "ReqIFText", optional=True
    )
    attributes = AttributeAccessor()
    relations = RequirementsRelationAccessor()
    related = ElementRelationAccessor()
    type = c.AttrProxyAccessor(RequirementType, "requirementType")


@c.xtype_handler(None, XT_FOLDER)
class RequirementsFolder(Requirement):
    """A folder that stores Requirements."""

    _xmltag = "ownedRequirements"

    folders: c.Accessor
    requirements = c.DirectProxyAccessor(
        Requirement, XT_REQUIREMENT, aslist=c.ElementList
    )


@c.xtype_handler(None, XT_MODULE)
class RequirementsModule(ReqIFElement):
    """A ReqIF Module that bundles multiple Requirement folders."""

    _xmltag = "ownedExtensions"

    folders = c.DirectProxyAccessor(
        RequirementsFolder, XT_FOLDER, aslist=c.ElementList
    )
    requirements = c.DirectProxyAccessor(
        Requirement, XT_REQUIREMENT, aslist=c.ElementList
    )
    type = c.AttrProxyAccessor(ModuleType, "moduleType")
    attributes = AttributeAccessor()

    def to_reqif(
        self,
        to: str | os.PathLike | t.IO[bytes],
        *,
        metadata: cabc.Mapping[str, t.Any] | None = None,
        pretty: bool = False,
        compress: bool | None = None,
    ) -> None:
        """Export this module as ReqIF XML.

        You can override some auto-generated metadata placed in the header
        section by passing a dictionary as ``metadata``. The following keys
        are supported. Unsupported keys are silently ignored.

        -   ``creation_time``: A ``datetime.datetime`` object that specifies
            this document's creation time. Default to the current time.
        -   ``comment``: Override the ReqIF file's comment. Defaults to a
            text derived from the model and module names.
        -   ``title``: Specify the document title. Defaults to the module's
            ``long_name``.

        Parameters
        ----------
        to
            Where to export to. Can be the name of a file, or a file-like
            object opened in binary mode.
        metadata
            A dictionary with additional metadata (see above).
        pretty
            Format the XML human-readable.
        compress
            Write compressed data (``*.reqifz``). Defaults to ``True``
            if ``target`` is a string or path-like and its name ends in
            ``.reqifz``, otherwise defaults to ``False``.
        """
        exporter.export_module(
            self, to, metadata=metadata, pretty=pretty, compress=compress
        )


class AbstractRequirementsRelation(ReqIFElement):
    _required_attrs = frozenset({"source", "target"})

    type = c.AttrProxyAccessor(RelationType, "relationType")
    source = c.AttrProxyAccessor(Requirement, "source")
    target = c.AttrProxyAccessor(c.GenericElement, "target")

    def _short_repr_(self) -> str:
        direction = ""
        if self.source is not None:
            direction += f" from {self.source._short_repr_()}"
        if self.target is not None:
            direction += f" to {self.target._short_repr_()}"
        return (
            f"<{type(self).__name__} {self.long_name!r}{direction} "
            f"({self.uuid})>"
        )


@c.xtype_handler(None, XT_OUT_RELATION)
class RequirementsOutRelation(AbstractRequirementsRelation):
    """A Relation between an object and a requirement."""

    _xmltag = "ownedExtensions"

    source = c.AttrProxyAccessor(Requirement, "target")
    target = c.AttrProxyAccessor(c.GenericElement, "source")


@c.xtype_handler(None, XT_INC_RELATION)
class RequirementsIncRelation(AbstractRequirementsRelation):
    """A Relation between a requirement and an object."""

    _xmltag = "ownedRelations"


@c.xtype_handler(None, XT_INT_RELATION)
class RequirementsIntRelation(AbstractRequirementsRelation):
    """A Relation between two requirements."""

    _xmltag = "ownedRelations"


@c.xtype_handler(None, XT_REQ_TYPES_F)
class RequirementsTypesFolder(ReqIFElement):
    _xmltag = "ownedExtensions"

    data_type_definitions = c.DirectProxyAccessor(
        c.GenericElement,
        (XT_REQ_TYPES_DATA_DEF, XT_REQ_TYPE_ENUM),
        aslist=c.MixedElementList,
    )
    module_types = c.DirectProxyAccessor(
        ModuleType, XT_MODULE_TYPE, aslist=c.ElementList
    )
    relation_types = c.DirectProxyAccessor(
        RelationType, XT_RELATION_TYPE, aslist=c.ElementList
    )
    requirement_types = c.DirectProxyAccessor(
        RequirementType, XT_REQ_TYPE, aslist=c.ElementList
    )


def init() -> None:
    c.set_accessor(
        RequirementsFolder,
        "folders",
        c.DirectProxyAccessor(
            RequirementsFolder, XT_FOLDER, aslist=c.ElementList
        ),
    )
    c.set_accessor(c.GenericElement, "requirements", ElementRelationAccessor())
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "requirement_modules",
        c.DirectProxyAccessor(
            RequirementsModule, XT_MODULE, aslist=c.ElementList
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_requirements",
        c.DeepProxyAccessor(
            Requirement,
            XT_REQUIREMENT,
            aslist=c.ElementList,
            rootelem=XT_MODULE,
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "requirement_types_folders",
        c.DirectProxyAccessor(
            RequirementsTypesFolder,
            XT_REQ_TYPES_F,
            aslist=c.ElementList,
        ),
    )
    c.set_accessor(
        RequirementsModule,
        "requirement_types_folders",
        c.DirectProxyAccessor(
            RequirementsTypesFolder,
            XT_REQ_TYPES_F,
            aslist=c.ElementList,
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_requirement_types",
        c.DeepProxyAccessor(
            RequirementType,
            XT_REQ_TYPE,
            aslist=c.ElementList,
            rootelem=XT_REQ_TYPES_F,
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_module_types",
        c.DeepProxyAccessor(
            ModuleType,
            XT_MODULE_TYPE,
            aslist=c.ElementList,
            rootelem=XT_REQ_TYPES_F,
        ),
    )
    c.set_accessor(
        crosslayer.BaseArchitectureLayer,
        "all_relation_types",
        c.DeepProxyAccessor(
            RelationType,
            XT_RELATION_TYPE,
            aslist=c.ElementList,
            rootelem=XT_REQ_TYPES_F,
        ),
    )


_attr_type_hints = {
    "int": (IntegerValueAttribute, XT_REQ_ATTR_INTEGERVALUE),
    "integer": (IntegerValueAttribute, XT_REQ_ATTR_INTEGERVALUE),
    "integervalueattribute": (IntegerValueAttribute, XT_REQ_ATTR_INTEGERVALUE),
    "str": (StringValueAttribute, XT_REQ_ATTR_STRINGVALUE),
    "string": (StringValueAttribute, XT_REQ_ATTR_STRINGVALUE),
    "stringvalueattribute": (StringValueAttribute, XT_REQ_ATTR_STRINGVALUE),
    "float": (RealValueAttribute, XT_REQ_ATTR_REALVALUE),
    "real": (RealValueAttribute, XT_REQ_ATTR_REALVALUE),
    "realvalueattribute": (RealValueAttribute, XT_REQ_ATTR_REALVALUE),
    "date": (DateValueAttribute, XT_REQ_ATTR_DATEVALUE),
    "datetime": (DateValueAttribute, XT_REQ_ATTR_DATEVALUE),
    "datevalueattribute": (DateValueAttribute, XT_REQ_ATTR_DATEVALUE),
    "bool": (BooleanValueAttribute, XT_REQ_ATTR_BOOLEANVALUE),
    "boolean": (BooleanValueAttribute, XT_REQ_ATTR_BOOLEANVALUE),
    "booleanvalueattribute": (BooleanValueAttribute, XT_REQ_ATTR_BOOLEANVALUE),
    "enum": (EnumerationValueAttribute, XT_REQ_ATTR_ENUMVALUE),
    "enumeration": (EnumerationValueAttribute, XT_REQ_ATTR_ENUMVALUE),
    "enumerationvalueattribute": (
        EnumerationValueAttribute,
        XT_REQ_ATTR_ENUMVALUE,
    ),
}



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Exceptions that may be raised by the PVMT module."""


class PropertyValueError(Exception):
    """Base class for all property-value related errors."""


class ScopeError(PropertyValueError):
    """Attempted to apply a PV group to an element outside its scope."""


class UndefinedKeyError(PropertyValueError, KeyError):
    """A key is attempted to be added which is not defined for the group."""


class CastingError(PropertyValueError):
    """A supplied value cannot be cast from or to the XML representation."""


class GroupNotAppliedError(PropertyValueError, KeyError):
    """The property value group has not been applied to this element."""
